package io.signpath.signpathclient;

import io.signpath.signpathclient.api.http.SignPathApiHttpClient;
import java.io.File;
import org.apache.commons.lang.Validate;
import io.signpath.signpathclient.api.model.SigningRequest;
import io.signpath.signpathclient.api.model.SigningRequestStatus;
import java.io.PrintStream;
import java.util.Map;
import java.util.concurrent.TimeUnit;

/**
 *
 * @author robert.benak
 */
public class SignPathClient {

    private final SignPathClientSimpleLogger logger;
    private final SignPathClientSettings signPathClientSettings;
    private final SignPathApiHttpClient signPathApiHttpClient;

    public SignPathClient(String signPathApiBaseUrl, SignPathClientSimpleLogger logger, SignPathClientSettings signPathClientSettings) {

        this.logger = logger;
        this.signPathClientSettings = signPathClientSettings;
        signPathApiHttpClient = new SignPathApiHttpClient(signPathApiBaseUrl, logger, signPathClientSettings);
    }

    public SignPathClient(String signPathApiBaseUrl, SignPathClientSimpleLogger logger) {
        this(signPathApiBaseUrl, logger, new SignPathClientSettings(
                (int) TimeUnit.MINUTES.toSeconds(10), // serviceUnavailableTimeoutInSeconds
                (int) TimeUnit.MINUTES.toSeconds(5), // uploadAndDownloadRequestTimeoutInSeconds
                (int) TimeUnit.MINUTES.toSeconds(10), // waitForCompletionTimeoutInSeconds
                (int) TimeUnit.SECONDS.toSeconds(30), // waitBetweenReadinessChecksInSeconds
                SignPathClient.buildUserAgent()
        ));
    }
    
    private static String buildUserAgent(){
        return String.format("SignPath.JavaClient/%1$s (OpenJDK %2$s)",
                SignPathClient.class.getPackage().getImplementationVersion(), System.getProperty("java.version"));
    }

    /**
     *
     * @param apiToken
     * @param organizationId
     * @param tbsToken
     * @param artifact
     * @param projectSlug
     * @param signingPolicySlug
     * @param artifactConfigurationSlug
     * @param description
     * @param origin
     * @return Signing Request Id
     */
    public String submitSigningRequest(
            String apiToken,
            String tbsToken,
            String organizationId,
            File artifact,
            String projectSlug,
            String signingPolicySlug,
            String artifactConfigurationSlug,
            String description,
            boolean allowIgnoreOrigin,
            Map<String, String> origin,
            Map<String, String> userDefinedParameters) throws SignPathClientException {

        Validate.notEmpty(apiToken, "SignPath API token is empty");
        Validate.notEmpty(tbsToken, "Trusted build system token is empty");
        Validate.notEmpty(organizationId, "Organization ID is empty");
        Validate.notNull(artifact, "Artifact is null");
        Validate.notEmpty(projectSlug, "Project slug is empty");
        Validate.notEmpty(signingPolicySlug, "Signing policy slug is empty");
        
        logger.log(String.format("[PARAM] serviceUnavailableTimeoutInSeconds: %s", this.signPathClientSettings.getServiceUnavailableTimeoutInSeconds()));
        logger.log(String.format("[PARAM] uploadAndDownloadRequestTimeoutInSeconds: %s", this.signPathClientSettings.getUploadAndDownloadRequestTimeoutInSeconds()));
        logger.log(String.format("[PARAM] waitForCompletionTimeoutInSeconds: %s", this.signPathClientSettings.getWaitForCompletionTimeoutInSeconds()));
        
        logger.log("The signing request is being submitted...");
        String signingRequestId = signPathApiHttpClient.submitSigningRequestHttp(
            apiToken, tbsToken, organizationId, artifact, projectSlug, signingPolicySlug,
            artifactConfigurationSlug, description, allowIgnoreOrigin, origin, userDefinedParameters);
        logger.log("The signing request has been successfully submitted");
        logger.log(String.format("The signing request id: %s", signingRequestId));
        return signingRequestId;
    }

    /**
     *
     * @param apiToken
     * @param tbsToken
     * @param organizationId
     * @param artifact
     * @param projectSlug
     * @param signingPolicySlug
     * @param artifactConfigurationSlug
     * @param artifactTargetFile
     * @param description
     * @param origin
     * @return
     * @throws java.lang.InterruptedException
     */
    public String submitSigningRequestAndWaitForSignedArtifact(
            String apiToken,
            String tbsToken,
            String organizationId,
            File artifact,
            String projectSlug,
            String signingPolicySlug,
            String artifactConfigurationSlug,
            File artifactTargetFile,
            String description,
            boolean allowIgnoreOrigin,
            Map<String, String> origin,
            Map<String, String> userDefinedParameters) throws SignPathClientException, InterruptedException {

        Validate.notEmpty(apiToken, "SignPath API token is empty");
        Validate.notEmpty(tbsToken, "Trusted build system token is empty");
        Validate.notEmpty(organizationId, "Organization ID is empty");
        Validate.notNull(artifact, "Artifact is null");
        Validate.notEmpty(projectSlug, "Project slug is empty");
        Validate.notNull(artifactTargetFile, "Artifact target file is null");
        Validate.notEmpty(signingPolicySlug, "Signing policy slug is empty");

        String signingRequestId = this.submitSigningRequest(
            apiToken, tbsToken, organizationId, artifact, projectSlug,
            signingPolicySlug, artifactConfigurationSlug, description, allowIgnoreOrigin,
            origin, userDefinedParameters);
        SigningRequest signingRequest = this.getSigningRequestWaitForFinalStatus(apiToken, organizationId, signingRequestId);
        SigningRequestStatus signingRequestStatus = signingRequest.getStatus();
        if (signingRequestStatus == SigningRequestStatus.Completed) {
            logger.log("Loading signed artifact...");
            // request is now approved and we can download the artifacts
            this.downloadSignedArtifact(apiToken, organizationId, signingRequestId, artifactTargetFile);
        }

        return signingRequestId;
    }

    /**
     * @param apiToken
     * @param organizationId
     * @param signingRequestId
     * @return
     */
    public SigningRequest getSigningRequest(
            String apiToken,
            String organizationId,
            String signingRequestId) {

        Validate.notEmpty(apiToken, "SignPath API token is empty");
        Validate.notEmpty(organizationId, "Organization ID is empty");
        Validate.notEmpty(signingRequestId, "Signing request ID is empty");

        SigningRequest signingRequest = signPathApiHttpClient.getSigningRequestHttp(apiToken, organizationId, signingRequestId);
        return signingRequest;
    }

    /**
     * @param apiToken
     * @param organizationId
     * @param signingRequestId
     * @return
     */
    public SigningRequest getSigningRequestWaitForFinalStatus(
            String apiToken,
            String organizationId,
            String signingRequestId) throws SignPathClientException {

        Validate.notEmpty(apiToken, "SignPath API token is empty");
        Validate.notEmpty(organizationId, "Organization ID is empty");
        Validate.notEmpty(signingRequestId, "Signing request ID is empty");

        long startTime = System.currentTimeMillis();

        logger.log("Checking the signing request status...");
        SigningRequest signingRequest = this.getSigningRequest(apiToken, organizationId, signingRequestId);
        int waitForCompletionTimeoutInSeconds = this.signPathClientSettings.getWaitForCompletionTimeoutInSeconds();
        while (!signingRequest.isFinalStatus() && System.currentTimeMillis() < startTime + waitForCompletionTimeoutInSeconds * 1000) {
            try {
                Thread.sleep(this.signPathClientSettings.getWaitBetweenReadinessChecksInSeconds() * 1000);
            } catch (InterruptedException ex) {
                throw new SignPathClientException("The waiting thread was interrupted");
            }
            logger.log("The signing request is not yet completed. Please complete the signing request in SignPath in order to proceed.");
            signingRequest = this.getSigningRequest(apiToken, organizationId, signingRequestId);
        }

        SigningRequestStatus signingRequestStatus = signingRequest.getStatus();
        if (signingRequestStatus != SigningRequestStatus.Completed) {
            // we were waiting for a while, but request is still not completed 
            switch (signingRequestStatus) {
                case Denied:
                    throw new SignPathClientException("Signing request was denied");
                case Canceled:
                    throw new SignPathClientException("Signing request was canceled");
                case Failed:
                    throw new SignPathClientException("Signing request failed");
                case InProgress:
                case WaitingForApproval:
                    throw new SignPathClientException("Timeout expired while waiting for signing request to complete");
                default:
                    throw new SignPathClientException(
                            String.format("Terminating because of unexpected signing request status: %s", signingRequestStatus));
            }
        }

        logger.log("The signing request is completed");
        return signingRequest;
    }

    /**
     *
     * @param apiToken
     * @param organizationId
     * @param signingRequestId
     * @param artifactTargetFile
     * @throws SignPathClientException
     */
    public void downloadSignedArtifact(
            String apiToken,
            String organizationId,
            String signingRequestId,
            File artifactTargetFile) throws SignPathClientException {

        Validate.notEmpty(apiToken, "SignPath API token is empty");
        Validate.notEmpty(organizationId, "Organization ID is empty");
        Validate.notEmpty(signingRequestId, "Signing request ID is empty");
        Validate.notNull(artifactTargetFile, "Artifact target file is null");

        signPathApiHttpClient.downloadSignedArtifactHttp(apiToken, organizationId, signingRequestId, artifactTargetFile);
    }
}
